@@ -1,11 +1,11 @@
 from typing import NamedTuple, Optional, Union
+from types import FunctionType
 from select import select
 from time import sleep, time
 from collections import deque
 from hashlib import sha256
 from binascii import a2b_hex
 from Cryptodome.Cipher import AES
-from Cryptodome.Util.Padding import pad, unpad
 from struct import Struct
 from io import BytesIO, SEEK_END
 from socket import socket
@@ -37,7 +37,7 @@ FLAG_NAMES = {
     CONTROL_MTU: "MTU",
     CONTROL_FIN: "FIN",
 }
-WINDOW_MAX_SIZE = 32768
+WINDOW_MAX_SIZE = 32768  # 32kb
 SEND_BUFFER_SIZE = WINDOW_MAX_SIZE * 8  # 256kb
 MAX_RETRANSMIT_LIMIT = 4
 FULL_SIZE_PACKET_WAIT = 0.001  # sec
@@ -101,6 +101,10 @@ class CycInt(int):
         return not self.__le__(other)
 
 
+# static cycle int
+CYC_INT0 = CycInt(0)
+
+
 class Packet(NamedTuple):
     """
     static 14b
@@ -144,7 +148,7 @@ class SecureReliableSocket(socket):
         "mut_auto_fix", "mtu_size", "mtu_multiple",
         "sender_seq", "sender_buffer", "sender_signal", "sender_buffer_lock",
         "receiver_seq", "receiver_buffer", "receiver_signal", "receiver_buffer_lock",
-        "loss", "established"]
+        "broadcast_hook_fnc", "loss", "established"]
 
     def __init__(self, family=s.AF_INET, timeout=21.0, span=3.0):
         """
@@ -171,6 +175,8 @@ class SecureReliableSocket(socket):
         self.receiver_buffer = BytesIO()
         self.receiver_signal = threading.Event()
         self.receiver_buffer_lock = threading.Lock()
+        # broadcast hook
+        self.broadcast_hook_fnc: Optional[FunctionType] = None
         # status
         self.loss = 0
         self.established = False
@@ -339,7 +345,7 @@ class SecureReliableSocket(socket):
 
             # connection may be broken
             if self.timeout < time() - last_receive_time:
-                p = Packet(CONTROL_FIN, CycInt(0), 0, time(), b'stream may be broken')
+                p = Packet(CONTROL_FIN, CYC_INT0, 0, time(), b'stream may be broken')
                 self.sendto(self._encrypt(packet2bin(p)), self.address)
                 break
 
@@ -389,7 +395,7 @@ class SecureReliableSocket(socket):
 
                 # receive reset
                 if packet.control & CONTROL_FIN:
-                    p = Packet(CONTROL_FIN, CycInt(0), 0, time(), b'be notified fin or reset')
+                    p = Packet(CONTROL_FIN, CYC_INT0, 0, time(), b'be notified fin or reset')
                     self.sendto(self._encrypt(packet2bin(p)), self.address)
                     break
 
@@ -431,7 +437,9 @@ class SecureReliableSocket(socket):
 
                 # broadcast packet
                 if packet.control & CONTROL_BCT:
-                    if last_packet is None or last_packet.control & CONTROL_EOF:
+                    if self.broadcast_hook_fnc is not None:
+                        self.broadcast_hook_fnc(packet)
+                    elif last_packet is None or last_packet.control & CONTROL_EOF:
                         self._push_receive_buffer(packet.data)
                     else:
                         broadcast_packets.append(packet)
@@ -586,7 +594,7 @@ class SecureReliableSocket(socket):
         # window_size = self.get_window_size()
         # if window_size < len(data):
         #    raise ValueError("data is too big {}<{}".format(window_size, len(data)))
-        packet = Packet(CONTROL_BCT, CycInt(0), 0, time(), data)
+        packet = Packet(CONTROL_BCT, CYC_INT0, 0, time(), data)
         with self.sender_buffer_lock:
             self.sendto(self._encrypt(packet2bin(packet)), self.address)
 
@@ -623,13 +631,15 @@ class SecureReliableSocket(socket):
         return b''
 
     def _encrypt(self, data: bytes) -> bytes:
+        """encrypt by AES-GCM (more secure than CBC mode)"""
         cipher = AES.new(self.shared_key, AES.MODE_GCM)
         # warning: Don't reuse nonce
         enc, tag = cipher.encrypt_and_digest(data)
-        # length = 16bytes + 16bytes + N(=data)bytes
+        # output length = 16bytes + 16bytes + N(=data)bytes
         return cipher.nonce + tag + enc
 
     def _decrypt(self, data: bytes) -> bytes:
+        """decrypt by AES-GCM (more secure than CBC mode)"""
         cipher = AES.new(self.shared_key, AES.MODE_GCM, nonce=data[:16])
         # ValueError raised when verify failed
         return cipher.decrypt_and_verify(data[32:], data[16:32])
@@ -643,7 +653,7 @@ class SecureReliableSocket(socket):
 
     def close(self) -> None:
         if self.established:
-            p = Packet(CONTROL_FIN, CycInt(0), 0, time(), b'closed')
+            p = Packet(CONTROL_FIN, CYC_INT0, 0, time(), b'closed')
             self.sendto(self._encrypt(packet2bin(p)), self.address)
             sleep(0.001)
             super().close()
@@ -671,7 +681,7 @@ def get_mtu_linux(family, host) -> int:
 
 def main():
     """for test"""
-    import sys
+    import sys, random
     remote_host = sys.argv[1]
     port = int(sys.argv[2])
     msglen = int(sys.argv[3])
@@ -695,9 +705,6 @@ def main():
             r = sock.recv(8192)
             if len(r) == 0:
                 break
-            if b'find me!' in r:
-                log.debug("find you!!! (%s)", r)
-                continue
             if 0 <= r.find(b'start!'):
                 size, start = 0, time()
             size += len(r)
@@ -711,9 +718,15 @@ def main():
         while msglen:
             sock.sendall(b'start!'+os.urandom(msglen)+b'success!')  # +14
             log.debug("send now! loss=%d time=%d", sock.loss, int(time()))
-            sock.broadcast(b'find me! ' + str(time()).encode())
+            if 0 == random.randint(0, 5):
+                sock.broadcast(b'find me! ' + str(time()).encode())
+                log.debug("send broadcast!")
             sleep(20)
 
+    def broadcast_hook(packet: Packet):
+        log.debug("find you!!! (%s)", packet)
+
+    sock.broadcast_hook_fnc = broadcast_hook
     threading.Thread(target=listen).start()
     threading.Thread(target=sending).start()
 
